
#import "WOStopWatch.h"
#import <Foundation/Foundation.h>
#import <Foundation/NSException.h>
#import <Foundation/NSString.h>
#if defined(__MACH__)
#import <libc.h>
#endif

NSTimeInterval _WOCurrentTime() {
#if defined(WIN32)
	return [[NSCalendarDate calendarDate] timeIntervalSince1970];
    //FILETIME ft;
    //double ti;
    //GetSystemTimeAsFileTime(&ft);
    //ti = (double)ft.dwHighDateTime * 4294967296.0 + (double)ft.dwLowDateTime;
    //ti -= 116444736000000000.0;	// 100-nanosecond units between 1601 and 1970
    //return (ti / 10000000.0) - NSTimeIntervalSince1970;
    //return 0;
#else
    struct timeval tv;
    gettimeofday(&tv, NULL);

#if defined(hpux)
// hpux gets confused because tv.tv_sec is an unsigned long, elsewhere it is just long
    {
    NSTimeInterval rv = tv.tv_sec;
    rv -= (long)NSTimeIntervalSince1970;
    rv += ( 1E-6 * (NSTimeInterval)tv.tv_usec );
    return rv;
    }
#else
    return (NSTimeInterval)(tv.tv_sec - (long)NSTimeIntervalSince1970) + 1E-6 * (NSTimeInterval)tv.tv_usec;
#endif
#endif
}

@implementation WOStopWatch : NSObject {
    NSTimeInterval _accumulatedTime;
    NSTimeInterval _currentStartTime;
    unsigned _startCount;
}

- (NSTimeInterval)accumulatedTime {
    return _accumulatedTime;
}

- (NSTimeInterval)currentRunningTime {
    NSTimeInterval aCurrentRunningTime = 0;
    if (_currentStartTime != 0) {
        aCurrentRunningTime = _WOCurrentTime() - _currentStartTime;
    } else {
        [NSException raise:NSGenericException format:@"%@: Attempt to access currentRunningTime when not running.", [self class]];
    }
    return aCurrentRunningTime;
}

- (void)start {
    if (_currentStartTime == 0) {
        _currentStartTime = _WOCurrentTime();
        _startCount++;
    } else {
        [NSException raise:NSGenericException format:@"%@: Attempt to start when already running.", [self class]];
    }
}

- (void)stop {
    NSTimeInterval aDeltaTime = [self currentRunningTime];
    _accumulatedTime += aDeltaTime;
    _currentStartTime = 0;
}

- (unsigned)startCount {
    return _startCount;
}

@end
