
#import "WXAssociation.h"
#import "WOStopWatch.h"
#import <WebObjects/WOApplication.h>
#import <WebObjects/WOComponent.h>
#import <WebObjects/WORequest.h>
#import <WebObjects/WOResponse.h>
#import <WebObjects/WOContext.h>
#import <WebObjects/WOSession.h>
#import <Foundation/NSString.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSEnumerator.h>
#import <Foundation/NSAutoreleasePool.h>
#import <Foundation/NSUserDefaults.h>
#import <Foundation/NSNotification.h>
#import <Foundation/NSThread.h>

#define WO_TIMING_CONTOL @"WOTimingControl"

@interface NSNumber (Extensions)
+ (NSNumber*)_sharedUnsignedInt:(unsigned)anUnsigned;
@end

@interface WOApplication (Private)
- (NSString*)_directConnectURL;
- (void)_openURL:(NSString*)aURL;
@end

@interface WOMultiKeyDictionary : NSObject{
}
- (id)objectForKeys:(NSString*)aKey,...;
- (void)setObject:(id)aValue forKeys:(NSString*)aKey,...;
- (NSEnumerator*)objectEnumerator;
- (void)removeAllObjects;
@end

@interface WOMultiKeyDictionary (Extensions)
- (NSArray*)allValues;
@end

@implementation WOMultiKeyDictionary (Extensions)

- (NSArray*)allValues {
    NSMutableArray* aMutableArray = [NSMutableArray array];
    NSEnumerator* anEnumerator = [self objectEnumerator];
    id aValue = nil;
    while (aValue = [anEnumerator nextObject]) {
        [aMutableArray addObject:aValue];
    }
    return aMutableArray;
}

@end

@interface NSString (WOFExtensions)
+ (NSString*)_stringWithStrings:(NSString*)aString, ...;
@end

static NSMutableArray* _requestResponseLoopTimings = nil;
static NSMutableDictionary* TheTimeIntervalDictionaryDictionary = nil;
static NSMutableDictionary* TheRequestResponseTimeIntervalDictionary = nil;
static unsigned TheRequestCount = 0;
static BOOL TheLoggingToConsoleFlag = NO;
static BOOL TheBindingTimingEnabledFlag = YES;
static unsigned TheDispatchRequestRepeatCount = 0;

@implementation WXAssociationTimeInterval : NSObject {
    WXKeyValueAssociation* _association;
    NSString* _componentName;
    WOStopWatch* _pushTimeStopWatch;
    WOStopWatch* _pullTimeStopWatch;
}

+ (NSDictionary*)pageTimingsDictionary {
    return TheTimeIntervalDictionaryDictionary;
}

+ (NSDictionary*)requestResponseTimings {
    return TheRequestResponseTimeIntervalDictionary;
}

+ (NSMutableDictionary *)createPhaseRecord:(NSString *)phaseName {
    NSMutableDictionary *phaseRecord = [NSMutableDictionary dictionary];
    NSMutableArray *timings = [NSMutableArray array];
    //total = 0.0;
    [phaseRecord setObject:[NSNumber _sharedUnsignedInt:0] forKey:@"total"];
    [phaseRecord setObject:phaseName forKey:@"name"];
    [phaseRecord setObject:timings forKey:@"array"];
    //[phaseRecord setObject:pageName forKey:@"pageName"];
    return phaseRecord;
}

+ (void)addRequestLoopTiming:(NSTimeInterval)interval phase:(NSString *)phaseName pageName:(NSString *)pageName {
    if (![pageName isEqualToString:WO_TIMING_CONTOL]) {
        NSMutableDictionary *phaseRecord;
        NSMutableArray *timings;
        NSTimeInterval total;
        NSMutableDictionary *pageRecord = [TheRequestResponseTimeIntervalDictionary objectForKey:pageName];
        if(!pageRecord){
            NSString * aPhaseName;
            pageRecord = [NSMutableDictionary dictionary];
            [pageRecord setObject: pageName forKey:@"name"];
            aPhaseName = @"takeValues";
            [pageRecord setObject: [self createPhaseRecord: aPhaseName] forKey: aPhaseName];
            aPhaseName = @"invokeAction";
            [pageRecord setObject: [self createPhaseRecord: aPhaseName] forKey: aPhaseName];
            aPhaseName = @"appendResponse";
            [pageRecord setObject: [self createPhaseRecord: aPhaseName] forKey: aPhaseName];
            [TheRequestResponseTimeIntervalDictionary setObject:pageRecord forKey:pageName];
        }
        phaseRecord = [pageRecord objectForKey: phaseName];
        if(!phaseRecord){
            // shouldn't ever get called
            [self createPhaseRecord: phaseName];
        }
        timings = [phaseRecord objectForKey: @"array"];
        total = [[phaseRecord objectForKey: @"total"] doubleValue];
        total = total + interval;
        [phaseRecord setObject:[NSNumber numberWithDouble:total] forKey:@"total"];
        [timings addObject:[NSNumber numberWithDouble:total]];
    }
}

+ (void)mark {
    [TheTimeIntervalDictionaryDictionary removeAllObjects];
    [TheRequestResponseTimeIntervalDictionary removeAllObjects];
}

+ (BOOL)bindingTimingEnabled {
    return TheBindingTimingEnabledFlag;
}

+ (void)setBindingTimingEnabled:(BOOL)aFlag {
    TheBindingTimingEnabledFlag = aFlag;
}

+ (unsigned)dispatchRequestRepeatCount {
    return TheDispatchRequestRepeatCount;
}

+ (void)setDispatchRequestRepeatCount:(int)anInt {
    TheDispatchRequestRepeatCount = (anInt >= 0) ? (unsigned)anInt : (unsigned)0;
    if (TheDispatchRequestRepeatCount > 30) {
        [WOApp setPageCacheSize:(TheDispatchRequestRepeatCount * 1.1)];
    }
}

+ (BOOL)loggingToConsole {
    return TheLoggingToConsoleFlag;
}

+ (void)setLoggingToConsole:(BOOL)aVal {
    TheLoggingToConsoleFlag = aVal;
}

+ (void)addRequestLoopTimingLog:(NSString *)aLog {
    [_requestResponseLoopTimings addObject:aLog];
}

+ (void)initialize {
    if (self == [WXAssociationTimeInterval class]) {
        NSLog(@"**** [WXApplication poseAsClass:[WOApplication class]] ****");
        [WXApplication poseAsClass:[WOApplication class]];

        NSLog(@"**** [WXKeyValueAssociation poseAsClass:[WOKeyValueAssociation class]] ****");
        [WXKeyValueAssociation poseAsClass:[WOKeyValueAssociation class]];

        NSLog(@"**** [WXBindingNameAssociation poseAsClass:[WOBindingNameAssociation class]] ****");
        [WXBindingNameAssociation poseAsClass:[WOBindingNameAssociation class]];

        //NSLog(@"**** [WXSession poseAsClass:[WOSession class]] ****");
        //[WXSession poseAsClass:[WOSession class]];

        TheTimeIntervalDictionaryDictionary = [[NSMutableDictionary alloc] init];
        _requestResponseLoopTimings = [[NSMutableArray array] retain];
        TheRequestResponseTimeIntervalDictionary = [[NSMutableDictionary dictionary] retain];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleApplicationDidDispatchRequestNotification:) name:@"WOApplicationDidDispatchRequestNotification" object:nil];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(handleApplicationDidFinishLaunchingNotification:) name:@"WOApplicationDidFinishLaunchingNotification" object:nil];
    }
}

+ (void)handleApplicationDidFinishLaunchingNotification:(NSNotification*)aNotification {
    NSString* aControlPageUrl = [WOApp _directConnectURL];
    aControlPageUrl = [aControlPageUrl stringByAppendingString:@".woa/wa/timingControl"];
    if ([[WOApp class] autoOpenInBrowser]) {
        [WOApp debugWithFormat:@"** Opening control page in browser at %@", aControlPageUrl];
        [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:5]];
        [WOApp _openURL:aControlPageUrl];
    } else {
        [WOApp debugWithFormat:@"** The control page can be acessed via %@", aControlPageUrl];
    }
}

+ (void)handleApplicationDidDispatchRequestNotification:(NSNotification*)aNotification {
    TheRequestCount++;
    [[self thisRequestResponseLoopTimings] removeAllObjects];
}

- (void)dealloc {
    [_association release];
    [_componentName release];
    [_pullTimeStopWatch release];
    [_pushTimeStopWatch release];
    [super dealloc];
}

- (id)initAssociation:(WOAssociation*)anAssociation componentName:(NSString*)aComponentName {
    self = [super init];
    _association = [anAssociation retain];
    _componentName = [aComponentName copy];
    _pushTimeStopWatch = [[WOStopWatch alloc] init];
    _pullTimeStopWatch = [[WOStopWatch alloc] init];
    return self;
}

- (void)startPullStopWatch {
    [_pullTimeStopWatch start];
}

- (void)stopPullStopWatch {
    [_pullTimeStopWatch stop];
}

- (void)startPushStopWatch {
    [_pushTimeStopWatch start];
}

- (void)stopPushStopWatch {
    [_pushTimeStopWatch stop];
}

- (NSString*)componentName {
    return _componentName;
}

- (NSString*)declarationName {
    return [_association declarationName];
}

- (NSString*)declarationType {
    return [_association declarationType];
}

- (NSString*)bindingName {
    return [_association bindingName];
}

- (NSString*)associationString {
    return [_association _keyPath];
}

- (NSTimeInterval)accumulatedPushTime {
    return [_pushTimeStopWatch accumulatedTime];
}

- (NSTimeInterval)accumulatedPullTime {
    return [_pullTimeStopWatch accumulatedTime];
}

- (NSTimeInterval)accumulatedTotalTime {
    return [self accumulatedPushTime] + [self accumulatedPullTime];
}

- (unsigned)usageCount {
    return [_pullTimeStopWatch startCount] + [_pushTimeStopWatch startCount];
}

- (NSComparisonResult)compareDescendingAccumulatedTotalTime:(WXAssociationTimeInterval*)anOtherAssociationTimeInterval {
    NSComparisonResult aComparisonResult;
    NSTimeInterval anAccumlatedTotalTimeInterval = [self accumulatedTotalTime];
    NSTimeInterval anOtherAccumlatedTotalTimeInterval = [anOtherAssociationTimeInterval accumulatedTotalTime];
    if (anOtherAccumlatedTotalTimeInterval < anAccumlatedTotalTimeInterval) {
        aComparisonResult = NSOrderedAscending;
    } else if (anOtherAccumlatedTotalTimeInterval > anAccumlatedTotalTimeInterval) {
        aComparisonResult = NSOrderedDescending;
    } else {
        aComparisonResult = NSOrderedSame;
    }
    return aComparisonResult;
}

- (NSComparisonResult)compareDescendingAccumulatedPushTime:(WXAssociationTimeInterval*)anOtherAssociationTimeInterval {
    NSComparisonResult aComparisonResult;
    NSTimeInterval anAccumlatedPushTimeInterval = [self accumulatedPushTime];
    NSTimeInterval anOtherAccumlatedPushTimeInterval = [anOtherAssociationTimeInterval accumulatedPushTime];
    if (anOtherAccumlatedPushTimeInterval < anAccumlatedPushTimeInterval) {
        aComparisonResult = NSOrderedAscending;
    } else if (anOtherAccumlatedPushTimeInterval > anAccumlatedPushTimeInterval) {
        aComparisonResult = NSOrderedDescending;
    } else {
        aComparisonResult = NSOrderedSame;
    }
    return aComparisonResult;
}

- (NSComparisonResult)compareDescendingAccumulatedPullTime:(WXAssociationTimeInterval*)anOtherAssociationTimeInterval {
    NSComparisonResult aComparisonResult;
    NSTimeInterval anAccumlatedPullTimeInterval = [self accumulatedPullTime];
    NSTimeInterval anOtherAccumlatedPullTimeInterval = [anOtherAssociationTimeInterval accumulatedPullTime];
    if (anOtherAccumlatedPullTimeInterval < anAccumlatedPullTimeInterval) {
        aComparisonResult = NSOrderedAscending;
    } else if (anOtherAccumlatedPullTimeInterval > anAccumlatedPullTimeInterval) {
        aComparisonResult = NSOrderedDescending;
    } else {
        aComparisonResult = NSOrderedSame;
    }
    return aComparisonResult;
}

- (NSComparisonResult)compareComponentDescendingAccumulatedTime:(WXAssociationTimeInterval*)anAssociationTimeInterval {
    NSComparisonResult aComparisonResult;
    NSString *aComponentName = [anAssociationTimeInterval componentName];
    aComparisonResult = [aComponentName caseInsensitiveCompare:[self componentName]];
    if(aComparisonResult == NSOrderedSame){
        aComparisonResult = [self compareDescendingAccumulatedTotalTime:anAssociationTimeInterval];
    }
    return aComparisonResult;
}

+ (NSArray*)thisRequestResponseLoopTimings {
    return _requestResponseLoopTimings;
}

+ (NSDictionary *)group:(NSArray *)theArray by:(NSString *)aKey {
    id aRecord, aVal, aInnerRecord;
    NSMutableArray * aArray;
    NSMutableDictionary * groupingDict;
    NSEnumerator * anEnum;
    double total;

    groupingDict = [NSMutableDictionary dictionary];
    anEnum = [theArray objectEnumerator];
    while(aRecord = [anEnum nextObject]) {
        aVal = [aRecord componentName];
        if(![groupingDict objectForKey:aVal]) {
            aArray = [NSMutableArray array];
            aInnerRecord = [NSMutableDictionary dictionary];
            total = 0.0;
            if(aVal){
                [aInnerRecord  setObject:aArray forKey:@"array"]; 
                [aInnerRecord  setObject:aVal forKey:@"name"]; 
                [aInnerRecord  setObject:[NSNumber numberWithDouble:total] forKey:@"total"]; 
                [groupingDict setObject:aInnerRecord forKey:aVal];
            }
        } else {
            aInnerRecord = [groupingDict objectForKey:aVal];
            aArray = [aInnerRecord objectForKey:@"array"];
            total = [[aInnerRecord objectForKey:@"total"] doubleValue];
        }
        [aArray addObject:aRecord];
        total = total + [aRecord accumulatedTotalTime];
        [aInnerRecord setObject:[NSNumber numberWithDouble:total] forKey:@"total"];
    }
    return groupingDict;
}

@end

static WXAssociationTimeInterval*
timeIntervalForAssociationWithBindingNameAndComponentNamed(WOAssociation* anAssociation, NSString* aDeclarationName, NSString* aBindingName, NSString* aComponentName, NSString* aPageName) {
    WXAssociationTimeInterval* anAssociationTimeInterval = nil;
    if (![aPageName isEqualToString:WO_TIMING_CONTOL]) {
        WOMultiKeyDictionary* aTimeIntervalDictionary = [TheTimeIntervalDictionaryDictionary objectForKey:aPageName];
        if (!aTimeIntervalDictionary) {
            aTimeIntervalDictionary = [[WOMultiKeyDictionary alloc] init];
            [TheTimeIntervalDictionaryDictionary setObject:aTimeIntervalDictionary forKey:aPageName];
            [aTimeIntervalDictionary release];
        }
        anAssociationTimeInterval = [aTimeIntervalDictionary objectForKeys:aBindingName, aDeclarationName, aComponentName, nil];
        if (!anAssociationTimeInterval) {
            anAssociationTimeInterval = [[WXAssociationTimeInterval alloc] initAssociation:anAssociation componentName:aComponentName];
            [aTimeIntervalDictionary setObject:anAssociationTimeInterval forKeys:aBindingName, aDeclarationName, aComponentName, nil];
            [anAssociationTimeInterval release];
        }
    }
    return anAssociationTimeInterval;
}

@implementation WXKeyValueAssociation

- (id)valueInComponent:(WOComponent*)aComponent {
    id aValue = nil;
    if (TheBindingTimingEnabledFlag) {
        NSString* aComponentName = [aComponent name];
        NSString* aPageName = [[[aComponent context] page] name];
        WXAssociationTimeInterval* anAssociationTimeInterval = timeIntervalForAssociationWithBindingNameAndComponentNamed(self, _bindingName, _declarationName, aComponentName, aPageName);
        [anAssociationTimeInterval startPullStopWatch];
        aValue = [super valueInComponent:aComponent];
        [anAssociationTimeInterval stopPullStopWatch];
    } else {
        aValue = [super valueInComponent:aComponent];        
    }
    return aValue;
}

- (void)setValue:(id)aValue inComponent:(WOComponent*)aComponent {
    if (TheBindingTimingEnabledFlag) {
        NSString* aComponentName = [aComponent name];
        NSString* aPageName = [[[aComponent context] page] name];
        WXAssociationTimeInterval* anAssociationTimeInterval = timeIntervalForAssociationWithBindingNameAndComponentNamed(self, _bindingName, _declarationName, aComponentName, aPageName);
        [anAssociationTimeInterval startPushStopWatch];
        [super setValue:aValue inComponent:aComponent];
        [anAssociationTimeInterval stopPushStopWatch];
    } else {
        [super setValue:aValue inComponent:aComponent];
    }
}

- (void)_setDebugEnabledForBinding:(NSString*)aBindingName declarationName:(NSString*)aDeclarationName declarationType:(NSString*)aDeclarationType {
    _bindingName = [aBindingName copy];
    _declarationName = [aDeclarationName copy];
    _declarationType = [aDeclarationType copy];
}

- (NSString*)declarationName {
    return _declarationName;
}

- (NSString*)declarationType {
    return _declarationType;
}

- (NSString*)bindingName {
    return _bindingName;
}

@end

@implementation WXBindingNameAssociation: WOBindingNameAssociation

- (id)valueInComponent:(WOComponent*)aComponent {
    id aValue = nil;
    if (TheBindingTimingEnabledFlag) {
        NSString* aComponentName = [aComponent name];
        NSString* aPageName = [[[aComponent context] page] name];
        WXAssociationTimeInterval* anAssociationTimeInterval = timeIntervalForAssociationWithBindingNameAndComponentNamed(self, _bindingName, _declarationName, aComponentName, aPageName);
        [anAssociationTimeInterval startPullStopWatch];
        aValue = [super valueInComponent:aComponent];
        [anAssociationTimeInterval stopPullStopWatch];
    } else {
        aValue = [super valueInComponent:aComponent];
    }
    return aValue;
}

- (void)setValue:(id)aValue inComponent:(WOComponent*)aComponent {
    if (TheBindingTimingEnabledFlag) {
        NSString* aComponentName = [aComponent name];
        NSString* aPageName = [[[aComponent context] page] name];
        WXAssociationTimeInterval* anAssociationTimeInterval = timeIntervalForAssociationWithBindingNameAndComponentNamed(self, _bindingName, _declarationName, aComponentName, aPageName);
        [anAssociationTimeInterval startPushStopWatch];
        [super setValue:aValue inComponent:aComponent];
        [anAssociationTimeInterval stopPushStopWatch];
    } else {
        [super setValue:aValue inComponent:aComponent];
    }
}

- (void)_setDebugEnabledForBinding:(NSString*)aBindingName declarationName:(NSString*)aDeclarationName declarationType:(NSString*)aDeclarationType {
    _bindingName = [aBindingName copy];
    _declarationName = [aDeclarationName copy];
    _declarationType = [aDeclarationType copy];
}

- (NSString*)declarationName {
    return _declarationName;
}

- (NSString*)declarationType {
    return _declarationType;
}

- (NSString*)bindingName {
    return _bindingName;
}

- (NSString*)_keyPath {
    return [NSString _stringWithStrings:@"^", _parentBindingName, (_keyPath ? @".":nil), _keyPath, nil];
}

@end

@interface WOAssociation (Private)
- (void)_setDebugEnabledForBinding:(NSString*)aBindingName declarationName:(NSString*)aDeclarationName declarationType:(NSString*)aDeclarationType;
@end

@implementation WOAssociation (Private)
- (void)_setDebugEnabledForBinding:(NSString*)aBindingName declarationName:(NSString*)aDeclarationName declarationType:(NSString*)aDeclarationType {
    // ** This prevents other assoc types from having WODebug accidently turned on.
}
@end

@interface WODeclaration : NSObject {
    NSString* _name;
    NSString* _type;
    NSDictionary* _associations;
}
@end

@implementation WODeclaration (Instrumentation)

- (void)_setDebuggingForAssociations {
    NSString* aBindingName = nil;
    NSEnumerator* aKeyEnumerator = [_associations keyEnumerator];
    while (aBindingName = [aKeyEnumerator nextObject]) {
        WOAssociation* anAssociation = [_associations objectForKey:aBindingName];
        [anAssociation _setDebugEnabledForBinding:aBindingName declarationName:_name declarationType:_type];
    }
}

@end

@implementation WODirectAction (TimingControl)

- (WOComponent *)timingControlAction {
    WOComponent *aPage;    
    // ** until we make this page all direct actions, force session timeout to be high for the control page.
    WOSession* aSession = [self session];
    [aSession setTimeOut:7200];

    aPage = [self pageWithName:WO_TIMING_CONTOL];
    return aPage;
}

@end

@implementation WXSession 

- (void)awake {
    NSDate* aDate = [[NSDate alloc] init];
    NSTimeInterval _baseRefInterval = [aDate timeIntervalSinceReferenceDate];
    [self setObject:[NSNumber numberWithDouble: _baseRefInterval] forKey:@"WXSession_baseRefInterval"];
    [aDate release];
}

- (void)takeValuesFromRequest:(WORequest *)r inContext:(WOContext *)c {
    NSDate* aDate = nil;
    NSTimeInterval aTimeInterval1, aTimeInterval2;
    NSTimeInterval _baseRefInterval = [[self objectForKey:@"WXSession_baseRefInterval"] doubleValue];

    aDate = [[NSDate alloc] init];
    aTimeInterval1 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    [[WXAssociationTimeInterval class] addRequestLoopTimingLog:[NSString stringWithFormat:@"session restore done: cum: %f   delta: %f", aTimeInterval1 - _baseRefInterval, aTimeInterval1 - _baseRefInterval]];
    [super takeValuesFromRequest:r inContext:c];
    aDate = [[NSDate alloc] init];
    aTimeInterval2 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    [[WXAssociationTimeInterval class] addRequestLoopTimingLog:[NSString stringWithFormat:@"takeValues done: cum: %f  delta: %f", aTimeInterval2 - _baseRefInterval, aTimeInterval2 - aTimeInterval1]];
    [[WXAssociationTimeInterval class] addRequestLoopTiming: aTimeInterval2 - aTimeInterval1 phase:@"takeValues" pageName:[[[self context] page] name]];
}

- (WOElement *)invokeActionForRequest:(WORequest *)r inContext:(WOContext *)c{
    WOElement *ele;
    NSDate* aDate = nil;
    NSTimeInterval aTimeInterval1, aTimeInterval2;
    NSTimeInterval _baseRefInterval = [[self objectForKey:@"WXSession_baseRefInterval"] doubleValue];

    aDate = [[NSDate alloc] init];
    aTimeInterval1 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    ele = [super invokeActionForRequest:r inContext:c];
    aDate = [[NSDate alloc] init];
    aTimeInterval2 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    [[WXAssociationTimeInterval class] addRequestLoopTimingLog:[NSString stringWithFormat:@"action invoke done: cum: %f  delta: %f", aTimeInterval2 - _baseRefInterval, aTimeInterval2 - aTimeInterval1]];
    [[WXAssociationTimeInterval class] addRequestLoopTiming: aTimeInterval2 - aTimeInterval1 phase:@"invokeAction" pageName:[[[self context] page] name]];
    return ele;
}

- (void)appendToResponse:(WOResponse*)aResponse inContext:(WOContext*)aContext {
    NSDate* aDate = nil;
    NSTimeInterval aTimeInterval1, aTimeInterval2;
    NSTimeInterval _baseRefInterval = [[self objectForKey:@"WXSession_baseRefInterval"] doubleValue];
    aDate = [[NSDate alloc] init];
    aTimeInterval1 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];

    [super appendToResponse:aResponse inContext:aContext];
    aDate = [[NSDate alloc] init];
    aTimeInterval2 = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    [self setObject:[NSNumber numberWithDouble:aTimeInterval2] forKey:@"WXSession_aTimeInterval2"];
    [[WXAssociationTimeInterval class] addRequestLoopTimingLog:[NSString stringWithFormat:@"append done: cum: %f  delta: %f", aTimeInterval2 - _baseRefInterval, aTimeInterval2 - aTimeInterval1]];
    [[WXAssociationTimeInterval class] addRequestLoopTiming: aTimeInterval2 - aTimeInterval1 phase:@"appendResponse" pageName:[[[self context] page] name]];
}

- (void)sleep{
    NSDate* aDate = nil;
    NSTimeInterval aTimeInterval = 0.0;
    NSTimeInterval _baseRefInterval = [[self objectForKey:@"WXSession_baseRefInterval"] doubleValue];
    NSTimeInterval _aTimeInterval2 = [[self objectForKey:@"WXSession_aTimeInterval2"] doubleValue];

    aDate = [[NSDate alloc] init];
    aTimeInterval = [aDate timeIntervalSinceReferenceDate];
    [aDate release];
    [[WXAssociationTimeInterval class] addRequestLoopTimingLog:[NSString stringWithFormat:@"session storage done: cum: %f  delta: %f", aTimeInterval - _baseRefInterval, aTimeInterval - _aTimeInterval2]];
}

@end

@implementation WXApplication : WOApplication

- (WOResponse*)dispatchRequest:(WORequest*)aRequest {
    WOResponse* aResponse = [super dispatchRequest:aRequest];
    unsigned aDispatchCount = (TheDispatchRequestRepeatCount > 0) ? TheDispatchRequestRepeatCount - 1 : 0;
    if (aDispatchCount > 0) {
        NSString* aHeader = [aResponse headerForKey:@"x-wxtimingcontrol-x"];
        if (!aHeader) {
            if (TheLoggingToConsoleFlag) {
                NSLog(@"** Internally executing -dispatchRequest: %d times.", aDispatchCount);
            }
            while (--aDispatchCount) {
                [super dispatchRequest:aRequest];
            }
            aResponse = [super dispatchRequest:aRequest];
        }
    }
    return aResponse;
}

@end
